///////////////////////////////////////////////////////////////////////////////
// File    : rectangle.h
// Content : basic rectangle structure
// History : 06-Feb-2006 - [rb] creation
//           10-Apr-2006 - [rb] rewrite
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

#if !defined(RECTANGLE_H)
#define RECTANGLE_H


/* includes ******************************************************************/

#include <cmath>
#include <ostream>


/* rectangle declaration *****************************************************/

struct Rectangle {
	typedef double value_type;
	value_type x0, y0, x1, y1;
	
	Rectangle(value_type x0 = value_type(0.), 
	          value_type y0 = value_type(0.),
	          value_type x1 = value_type(1.), 
			  value_type y1 = value_type(1.)) :
		x0(x0), y0(y0), x1(x1), y1(y1) {
	}
		
	// accessors
	inline value_type w() const { return x1 - x0; }
	inline value_type h() const { return y1 - y0; }
	inline value_type size() const { return w()*h(); }
	inline value_type depth() const { 
		using namespace std;
		return log2(size()); 
	}
		
	// manipulators
	inline Rectangle &crop(const Rectangle &r = Rectangle()) {
		x0 = std::max(x0, r.x0); x1 = std::min(x1, r.x1);
		y0 = std::max(y0, r.y0); y1 = std::min(y1, r.y1); 
		return *this;
	}

	inline Rectangle &dilate(value_type dx, value_type dy) {
		dx = std::max(dx, -w()/2);
		dy = std::max(dy, -h()/2);
		x0 -= dx; x1 += dx;
		y0 -= dy; y1 += dy;
		return *this;
	}

	// predicates
	inline bool contains(value_type x, value_type y) const {
		return x0 <= x && x <= x1 && y0 <= y && y <= y1; }
	inline bool intercepts(const Rectangle &r) const {
		return !disjoint(*this, r); }
	
	friend inline bool disjoint(const Rectangle &p, const Rectangle &q) {
		return p.x1 < q.x0 || q.x1 < p.x0 || p.y1 < q.y0 || q.y1 < p.y0; }

	friend std::ostream &operator<<(std::ostream &os, const Rectangle &r) {
		return os << "((" << r.x0 << ", " << r.y0 << "),"
		          << " (" << r.x1 << ", " << r.y1 << "))"; }
};


#endif // !defined(RECTANGLE_H)
