///////////////////////////////////////////////////////////////////////////////
// File    : renderer.cpp
// Content : treemap renderer
// History : 10-Apr-2006 - [rb] creation
//           19-Oct-2007 - [rb] added CeCIIL licence, prepared distro
///////////////////////////////////////////////////////////////////////////////

/* licence ********************************************************************
Copyright Renaud Blanch [http://iihm.imag.fr/blanch/]

This software is a computer program whose purpose is to explore new interaction
techniques for zoomable treemaps.

This software is governed by the CeCILL license under French law and abiding by 
the rules of distribution of free software. You can  use, modify and/ or 
redistribute the software under the terms of the CeCILL license as circulated  
at the following URL [http://www.cecill.info]. 

As a counterpart to the access to the source code and  rights to copy, modify 
and redistribute granted by the license, users are provided only with a limited 
warranty  and the software's author,  the holder of the economic rights,  and 
the successive licensors  have only limited liability. 

The fact that you are presently reading this means that you have had knowledge 
of the CeCILL license and that you accept its terms.
******************************************************************************/

/* includes ******************************************************************/

#include "renderer.h"

#include <cmath>
#include <cassert>
using namespace std;

#if defined(__APPLE__)
#	include <GLUT/glut.h>
#else
#	include <GL/glut.h>
#endif

#include "colors.h"
#include "rectangle.h"
#include "view.h"


/* static data initialisation ************************************************/

//double Renderer::depth_max = -20;
double Renderer::depth_max = -10;
int Renderer::start_time;	
double Renderer::ptx = 0.;
double Renderer::pty = 0.;


/* factory *******************************************************************/

template< char c = 'o' > struct QuadType        { typedef OutlineQuad Q; };
template< >              struct QuadType< 'r' > { typedef RampQuad    Q; };

template< char c = 'o' > struct TextType        { typedef OrthogonalText T; };
template< >              struct TextType< 'h' > { typedef HorizontalText T; };
template< >              struct TextType< 'c' > { typedef CenteredText   T; };


Renderer *Renderer::create(char q, char t, const View *v, 
	                       int t_min, int t_max, float a, int p) {
	#define word(quad, text) (quad << 8 | text)
	#define case(quad, text) { \
	case word(quad, text): \
		return new CRenderer< QuadType< quad >::Q, \
		                      TextType< text >::T >(v, t_min, t_max, a, p); \
		break; \
	}

	switch(word(q, t)) {
		case('o', 'o');
		case('o', 'h');
		case('o', 'c');
		case('r', 'o');
		case('r', 'h');
		case('r', 'c');

	default:
		return new CRenderer< QuadType< >::Q, 
		                      TextType< >::T >(v, t_min, t_max, a, p);
		break;
	}
	#undef word
	#undef case
}


/* Graphical Renderer implementation *****************************************/

Renderer::Renderer(const View *v, int t_min, int t_max, float a, int p) : 
	view(v), text_size_min(t_min), text_size_max(t_max), alpha(a), palette(p) {
	
	// opengl state initialization
	glClearColor(1, 1, 1, 1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_LINE_SMOOTH);
}

void Renderer::preamble() const {
	start_time = glutGet(GLUT_ELAPSED_TIME); 
	ptx = view->sx();
	pty = view->sy();
	depth_max  = min(depth_max, view->box.depth());
	depth_max  = max(depth_max, log2(ptx*pty));
}

void Renderer::postamble() const {
	static bool fast = false; // store if previous display was fast
	
	float fps = 1000. / (glutGet(GLUT_ELAPSED_TIME) - start_time + 1);
	
	bool redisplay = false;
	if(fps < 10) {
		// slow
		if(!fast) {
			// avoid unstable transitions dirsect from fast to slow
			redisplay = true;
			depth_max += (10 - fps)*2;
		}
		fast = false;
		
	} else
	if(fps < 20) {
		// normal
		fast = false;
		
	} else {
		// fast
		depth_max -= .2;
		redisplay = true;
		fast = true;
	}
	
	if(redisplay && 
	   (view->getAnimationState() == 0) && 
	   (depth_max >= log2(ptx*pty))) {
		glutPostRedisplay();
	}
}

bool Renderer::top_down(const Rectangle &box) const {
	return box.intercepts(view->box) && box.depth() >= depth_max;
}

void Renderer::bottom_up(const Rectangle &box, int level, 
                          const std::string &label) const {
	glPushMatrix();
	glTranslatef(0, 0, box.depth());
	quad(box, level);
	text(box, level, label);
	glPopMatrix();
}


/* concrete quad renderers ***************************************************/

void OutlineQuad::quad(const Rectangle &box, int level) const {
	glColor4f(1, 1, 1, alpha);
	glBegin(GL_QUADS);
	glVertex3f(box.x0, box.y0, .5);
	glVertex3f(box.x0, box.y1, .5);
	glVertex3f(box.x1, box.y1, .5);
	glVertex3f(box.x1, box.y0, .5);
	glEnd();
	glColor3fv(colors::palettes[palette][level % 4]);
	glBegin(GL_LINE_LOOP);
	glVertex3f(box.x0, box.y0, -1);
	glVertex3f(box.x0, box.y1, -1);
	glVertex3f(box.x1, box.y1, -1);
	glVertex3f(box.x1, box.y0, -1);
	glEnd();	
}


static const float light  = 1.;
static const float dark   = .8;
static const float middle = (light + dark)/2;

RampQuad::RampQuad() {
	glClearColor(middle, middle, middle, 1.);
}

void RampQuad::quad(const Rectangle &box, int level) const {	
	glBegin(GL_QUADS);
	glColor4f(dark, dark, dark, alpha);
	glVertex3f(box.x0, box.y0, .5);
	glColor4f(middle, middle, middle, alpha);
	glVertex3f(box.x0, box.y1, .5);
	glColor4f(light, light, light, alpha);
	glVertex3f(box.x1, box.y1, .5);
	glColor4f(middle, middle, middle, alpha);
	glVertex3f(box.x1, box.y0, .5);
	glEnd();
	glColor3fv(colors::palettes[palette][level % 4]);
	glBegin(GL_LINE_LOOP);
	glVertex3f(box.x0, box.y0, -1);
	glVertex3f(box.x0, box.y1, -1);
	glVertex3f(box.x1, box.y1, -1);
	glVertex3f(box.x1, box.y0, -1);
	glEnd();	
}


/* concrete text renderers ***************************************************/

// globals
static const double TEXT_Z_OFFSET = .1;
static const double TEXT_SIZE     = 104.76; // glut font size (according to man)
static const double TEXT_UNIT     = 1./TEXT_SIZE;
static const int    TEXT_MARGIN   = 1; // 5;


void HorizontalText::text(const Rectangle &box, int level, 
                          const std::string &label) const {
	Rectangle cbox(box);
//	cbox.crop(view->box).dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	cbox.dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	
	// size in pixels
	double W = cbox.w() / ptx;
	double H = cbox.h() / pty;

	// text size
	double text_size = min(W / label.length(), H / 1.5);

	// too small
	if(text_size < text_size_min) return;
	if(text_size > text_size_max) text_size = text_size_max;

	// text top left corner & offset
	double cx = cbox.x0;
	double cy = cbox.y0;
	double tx = 0;
	double ty = 0.35;
	
	if(level % 2 == 1) {
		cy = cbox.y1;
		ty = -1.25;
	}

	glTranslatef(cx, cy, -TEXT_Z_OFFSET);
	glScalef(text_size * ptx, text_size * pty, 1.);
	glTranslatef(tx, ty, 0);
	glScalef(TEXT_UNIT, TEXT_UNIT, 1.);

	int line_width = int(text_size/20) + 1;

	glPushMatrix();
	glLineWidth(line_width + 2);
	glColor4f(1, 1, 1, .8);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glPopMatrix();

	glLineWidth(line_width);
	glColor3fv(colors::palettes[palette][level % 4]);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glLineWidth(1);
}

void CenteredText::text(const Rectangle &box, int level, 
                          const std::string &label) const {
	Rectangle cbox(box);
//	cbox.crop(view->box).dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	cbox.dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	
	// size in pixels
	double W = cbox.w() / ptx;
	double H = cbox.h() / pty;

	// text size
	double text_size = min(W / label.length(), H / 1.5);

	// too small
	if(text_size < text_size_min) return;
	if(text_size > text_size_max) text_size = text_size_max;

	// text top left corner & offset
	double cx = (cbox.x0 + cbox.x1) / 2;
	double cy = (cbox.y0 + cbox.y1) / 2;
	double tx = -.5 * label.length();
	double ty = -.5;

	glTranslatef(cx, cy, -TEXT_Z_OFFSET);
	glScalef(text_size * ptx, text_size * pty, 1.);
	glTranslatef(tx, ty, 0);
	glScalef(TEXT_UNIT, TEXT_UNIT, 1.);

	int line_width = int(text_size/20) + 1;

	glPushMatrix();
	glLineWidth(line_width + 2);
	glColor4f(1, 1, 1, .8);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glPopMatrix();

	glLineWidth(line_width);
	glColor3fv(colors::palettes[palette][level % 4]);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glLineWidth(1);
}

void OrthogonalText::text(const Rectangle &box, int level, 
                          const std::string &label) const {
	Rectangle cbox(box);
//	cbox.crop(view->box).dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	cbox.dilate(-TEXT_MARGIN*ptx, -TEXT_MARGIN*pty);
	
	// size in pixels
	double W = cbox.w() / ptx;
	double H = cbox.h() / pty;

	// text size
	bool horizontal = (level % 2) != 0;
	double lw = (horizontal ? label.length() : 1.5);
	double lh = (horizontal ? 1.5 : label.length());

	double text_size = min(W / lw, H / lh);

	// too small
	if(text_size < text_size_min) return;
	if(text_size > text_size_max) text_size = text_size_max;

	// text top left corner & offset
	double cx = cbox.x0;
	double cy = cbox.y0;
	double tx = 0;
	double ty = 0.35;
	
	switch(level % 4) {
	case 0:
	case 2:
		cx = cbox.x1;
		break;
		
	case 1:
		cy = cbox.y1;
		ty = -1.25;
		break;

	case 3:
		break;

	default:
		assert(false);
		break;
	}
	
	glTranslatef(cx, cy, -TEXT_Z_OFFSET);
	glScalef(text_size * ptx, text_size * pty, 1.);
	if(not horizontal) glRotatef(90, 0, 0, 1);
	glTranslatef(tx, ty, 0);
	glScalef(TEXT_UNIT, TEXT_UNIT, 1.);

	int line_width = int(text_size/20) + 1;

	glPushMatrix();
	glLineWidth(line_width + 2);
	glColor4f(1, 1, 1, .8);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glPopMatrix();

	glLineWidth(line_width);
	glColor3fv(colors::palettes[palette][level % 4]);
	for(const char *c = label.c_str(); *c != 0; ++c) {
		glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *c);
	}
	glLineWidth(1);
}
